import { cn } from "@/lib/utils";
import { getTagKind, tagBadgeVariants } from "./TagBadge";
import { useEffect, useRef, useState } from "react";
import { Button } from "../ui/button";
import { useOnClickOutside } from "usehooks-ts";
import { useTags } from "@/hook/useTags";
import { Skeleton } from "../ui/skeleton";

interface TagsSelectListProps {
    className?: string;
    search?: string;
    onSelect?: (tag: string) => void;
    onFirstArrowUp?: () => void;
}

function TagsSelectList({
    search = "",
    onSelect,
    onFirstArrowUp = () => { },
}: TagsSelectListProps) {
    const { data, isLoading } = useTags();
    const candidates = data?.filter(s => s.name.startsWith(search));

    return <ul className="max-h-[400px] overflow-scroll overflow-x-hidden">
        {isLoading && <>
            <li><Skeleton /></li>
            <li><Skeleton /></li>
            <li><Skeleton /></li>
        </>}
        {
            candidates?.length === 0 && <li className="p-2">No results</li>
        }
        {candidates?.map((tag) => <li key={tag.name}
            className="hover:bg-accent cursor-pointer p-1 rounded-sm transition-colors
            focus:outline-none focus:bg-accent focus:text-accent-foreground"
            tabIndex={-1}
            onClick={() => onSelect?.(tag.name)}
            onKeyDown={(e) => {
                if (e.key === "Enter") {
                    onSelect?.(tag.name);
                }
                if (e.key === "ArrowDown") {
                    const next = e.currentTarget.nextElementSibling as HTMLElement;
                    next?.focus();
                    e.preventDefault();
                }
                if (e.key === "ArrowUp") {
                    const prev = e.currentTarget.previousElementSibling as HTMLElement;
                    if (prev){
                        prev.focus();
                    }
                    else {
                        onFirstArrowUp();
                    }
                    e.preventDefault();
                }
            }}
            onPointerMove={(e) => {
                e.currentTarget.focus();
            }}
        >{tag.name}</li>)}
    </ul>
}

interface TagInputProps {
    className?: string;
    tags: string[];
    onTagsChange: (tags: string[]) => void;
    input: string;
    onInputChange: (input: string) => void;
}

export default function TagInput({
    className,
    tags = [],
    onTagsChange = () => { },
    input = "",
    onInputChange = () => { },
}: TagInputProps) {
    const inputRef = useRef<HTMLInputElement>(null);
    const setTags = onTagsChange;
    const setInput = onInputChange;
    const [isFocused, setIsFocused] = useState(false);
    const [openInfo, setOpenInfo] = useState<{
        top: number;
        left: number;
    } | null>(null);
    const autocompleteRef = useRef<HTMLDivElement>(null);
    useOnClickOutside(autocompleteRef, () => {
        setOpenInfo(null);
    });
    useEffect(() => {
        const listener = (e: KeyboardEvent) => {
            if (e.key === "Escape") {
                setOpenInfo(null);
            }
        }
        document.addEventListener("keyup", listener);
        return () => {
            document.removeEventListener("keyup", listener);
        }
    }, []);

    return <>
        {/* biome-ignore lint/a11y/useKeyWithClickEvents: input exist */}
        <div className={cn(`flex h-9 w-full rounded-md border border-input bg-transparent
        px-3 py-1 text-sm shadow-sm transition-colors justify-start items-center pr-0
        focus-visible:outline-none focus-visible:ring-1 focus-visible:ring-ring 
        disabled:cursor-not-allowed disabled:opacity-50`,
            isFocused && "outline-none ring-1 ring-ring",
            className
        )}
            onClick={() => inputRef.current?.focus()}
        >
            <ul className="flex gap-1 flex-none">
                {tags.map((tag) => <li className={cn(
                    tagBadgeVariants({ variant: getTagKind(tag) }),
                    "cursor-pointer"
                )} key={tag} onPointerDown={() =>{
                    setTags(tags.filter(x=>x!==tag));
                }}>{tag}</li>)}
            </ul>
            <input ref={inputRef} type="text" className="flex-1 border-0 ml-2 focus:border-0 focus:outline-none
            bg-transparent text-sm" placeholder="Add tag"
                onFocus={() => setIsFocused(true)} onBlur={() => setIsFocused(false)}
                value={input} onChange={(e) => setInput(e.target.value)} onKeyDown={(e) => {
                    if (e.key === "Enter") {
                        if (input.trim() === "") return;
                        setTags([...tags, input]);
                        setInput("");
                        setOpenInfo(null);
                    }
                    if (e.key === "Backspace" && input === "") {
                        setTags(tags.slice(0, -1));
                        setOpenInfo(null);
                    }
                    if (e.key === ":" || (e.ctrlKey && e.key === " ")) {
                        if (inputRef.current) {
                            const rect = inputRef.current.getBoundingClientRect();
                            setOpenInfo({
                                top: rect.bottom,
                                left: rect.left,
                            });
                        }
                    }
                    if (e.key === "Down" || e.key === "ArrowDown") {
                        if (openInfo && autocompleteRef.current) {
                            const firstChild = autocompleteRef.current.firstElementChild?.firstElementChild as HTMLElement;
                            firstChild?.focus();
                            e.preventDefault();
                        }
                    }
                }}
            />
            {
                openInfo && <div
                    ref={autocompleteRef}
                    className="absolute z-10 shadow-md bg-popover text-popover-foreground
                    border
                    rounded-sm p-2 w-[200px]"
                    style={{ top: openInfo.top, left: openInfo.left }}
                >
                    <TagsSelectList search={input} onSelect={(tag) => {
                        setTags([...tags, tag]);
                        setInput("");
                        setOpenInfo(null);
                    }} 
                    onFirstArrowUp={() => {
                        inputRef.current?.focus();
                    }}
                    />
                </div>
            }
            {
                tags.length > 0 && <Button variant="ghost" className="flex-none" onClick={() => {
                    setTags([]);
                    setOpenInfo(null);
                }}>Clear</Button>
            }
        </div>
    </>
}