import { useEffect, useReducer, useState } from "react";

interface AtomState<T> {
    value: T;
    listeners: Set<() => void>;
}
interface Atom<T> {
    key: string;
    default: T;
}

const atomStateMap = new WeakMap<Atom<unknown>, AtomState<unknown>>();

export function atom<T>(key: string, defaultVal: T): Atom<T> {
    return { key, default: defaultVal };
}

function getAtomState<T>(atom: Atom<T>): AtomState<T> {
    let atomState = atomStateMap.get(atom);
    if (!atomState) {
        atomState = {
            value: atom.default,
            listeners: new Set(),
        };
        atomStateMap.set(atom, atomState);
    }
    return atomState as AtomState<T>;
}

export function useAtom<T>(atom: Atom<T>): [T, (val: T) => void] {
    const state = getAtomState(atom);
    const [, setState] = useState(state.value);
    useEffect(() => {
        const listener = () => setState(state.value);
        state.listeners.add(listener);
        return () => {
            state.listeners.delete(listener);
        };
    }, [state]);
    return [
        state.value as T,
        (val: T) => {
            state.value = val;
            // biome-ignore lint/complexity/noForEach: forEach is used to call each listener
            state.listeners.forEach((listener) => listener());
        },
    ];
}

export function useAtomValue<T>(atom: Atom<T>): T {
    const state = getAtomState(atom);
    const update = useReducer((x) => x + 1, 0)[1];
    useEffect(() => {
        const listener = () => update();
        state.listeners.add(listener);
        return () => {
            state.listeners.delete(listener);
        };
    }, [state, update]);
    return state.value;
}

export function setAtomValue<T>(atom: Atom<T>): (val: T) => void {
    const state = getAtomState(atom);
    return (val: T) => {
        state.value = val;
        // biome-ignore lint/complexity/noForEach: forEach is used to call each listener
        state.listeners.forEach((listener) => listener());
    };
}