import { Card, CardContent, CardDescription, CardHeader, CardTitle } from "@/components/ui/card";
import { useGalleryDoc } from "../hook/useGalleryDoc.ts";
import TagBadge from "@/components/gallery/TagBadge";
import StyledLink from "@/components/gallery/StyledLink";
import { cn } from "@/lib/utils";
import { Link } from "wouter";

export interface ContentInfoPageProps {
    params: {
        id: string;
    };
}

interface TagClassifyResult {
    artist: string[];
    group: string[];
    series: string[];
    type: string[];
    character: string[];
    rest: string[];
}

function classifyTags(tags: string[]): TagClassifyResult {
    const result = {
        artist: [],
        group: [],
        series: [],
        type: [],
        character: [],
        rest: [],
    } as TagClassifyResult;
    const tagKind = new Set(["artist", "group", "series", "type", "character"]);
    for (const tag of tags) {
        const split = tag.split(":");
        if (split.length !== 2) {
            continue;
        }
        const [prefix, name] = split;
        if (tagKind.has(prefix)) {
            result[prefix as keyof TagClassifyResult].push(name);
        } else {
            result.rest.push(tag);
        }
    }
    return result;
}

export function ContentInfoPage({ params }: ContentInfoPageProps) {
    const { data, error, isLoading } = useGalleryDoc(params.id);

    if (isLoading) {
        return <div className="p-4">Loading...</div>
    }

    if (error) {
        return <div className="p-4">Error: {String(error)}</div>
    }

    if (!data) {
        return <div className="p-4">Not found</div>
    }

    const tags = data?.tags ?? [];
    const classifiedTags = classifyTags(tags);

    const contentLocation = `/doc/${params.id}/reader`;

    return (
        <div className="p-4">
            <Link to={contentLocation}>
                <div className="m-auto h-[400px] mb-4 flex justify-center items-center flex-none bg-[#272733]
            rounded-xl shadow-lg overflow-hidden">
                    <img
                        className="max-w-full max-h-full object-cover object-center"
                        src={`/api/doc/${data.id}/comic/thumbnail`}
                        alt={data.title} />
                </div>
            </Link>
            <Card className="flex-1">
                <CardHeader>
                    <CardTitle>
                        <StyledLink to={contentLocation}>
                            {data.title}
                        </StyledLink>
                    </CardTitle>
                    <CardDescription>
                        <StyledLink to={`/search?allow_tag=type:${classifiedTags.type[0] ?? ""}`} className="text-sm">
                            {classifiedTags.type[0] ?? "N/A"}
                        </StyledLink>
                    </CardDescription>
                </CardHeader>
                <CardContent>
                    <div className="grid gap-y-4 gap-x-3 lg:grid-cols-2">
                        <DescTagItem name="artist" items={classifiedTags.artist} />
                        <DescTagItem name="group" items={classifiedTags.group} />
                        <DescTagItem name="series" items={classifiedTags.series} />
                        <DescTagItem name="character" items={classifiedTags.character} />
                        <DescItem name="Created At">{new Date(data.created_at).toLocaleString()}</DescItem>
                        <DescItem name="Modified At">{new Date(data.modified_at).toLocaleString()}</DescItem>
                        <DescItem name="Path">{`${data.basepath}/${data.filename}`}</DescItem>
                        <DescItem name="Page Count">{JSON.stringify(data.additional)}</DescItem>
                    </div>
                    <div className="grid mt-4">
                        <span className="text-muted-foreground text-sm">Tags</span>
                        <ul className="mt-2 flex flex-wrap gap-1">
                            {classifiedTags.rest.map((tag) => <TagBadge key={tag} tagname={tag} />)}
                        </ul>
                    </div>
                </CardContent>
            </Card>
        </div>
    );
}

export default ContentInfoPage;

function DescItem({ name, children, className }: {
    name: string,
    className?: string,
    children?: React.ReactNode
}) {
    return <div className={cn("grid content-start", className)}>
        <span className="text-muted-foreground text-sm">{name}</span>
        <span className="text-primary leading-4 font-medium">{children}</span>
    </div>;
}

function DescTagItem({
    items,
    name,
    className,
}: {
    name: string;
    items: string[];
    className?: string;
}) {
    return <DescItem name={name} className={className}>
        {items.length === 0 ? "N/A" : items.map(
            (x) => <StyledLink key={x} to={`/search?allow_tag=${name}:${x}`}>{x}</StyledLink>
        )}
    </DescItem>
}
